"""
Procuret API
Email Address Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.time.time import ProcuretTime


class EmailAddress(Codable):

    def __init__(
        self,
        created: ProcuretTime,
        email_address: str,
        confirmation_required: bool
    ) -> None:

        self._created = created
        self._email_address = email_address
        self._confirmation_required = confirmation_required

        return
    
    created = property(lambda s: s._created)
    email_address = property(lambda s: s._email_address)
    confirmation_required = property(lambda s: s._confirmation_required)



"""
Procuret Python
Instalment Link Module
author: hugh@blinkybeach.com
"""
from xmlrpc.client import Boolean
from procuret.ancillary.communication_option import CommunicationOption
from typing import TypeVar, Type, Union, List
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.ancillary.entity_headline import EntityHeadline
from procuret.errors.type_error import ProcuretTypeError
from procuret.http.api_request import ApiRequest, HTTPMethod
from procuret.errors.inconsistent import InconsistentState
from procuret.session import Session
from decimal import Decimal
from enum import Enum
from procuret.data.order import Order
from procuret.http.query_parameters import QueryParameter, QueryParameters
from typing import Optional
from procuret.instalment_link.open import InstalmentLinkOpen
from procuret.time.time import ProcuretTime
from procuret.ancillary.sale_nomenclature import SaleNomenclature
from procuret.money.currency import Currency
from procuret.money.amount import Amount


Self = TypeVar('Self', bound='InstalmentLink')


class OrderBy(Enum):
    CREATED = 'created'


class InstalmentLink(Codable):

    path = '/instalment-link'
    list_path = path + '/list'

    _LINK_TEMPLATE = 'https://procuret.com/b?il={public_id}'

    coding_map = {
        'public_id': CD(str),
        'created': CD(ProcuretTime),
        'supplier': CD(EntityHeadline),
        'invoice_amount': CD(Decimal),
        'invitee_email': CD(str, optional=True),
        'invitee_phone_digits': CD(str, optional=True),
        'invitee_human_id': CD(int, optional=True),
        'invoice_identifier': CD(str),
        'opens': CD(InstalmentLinkOpen, array=True),
        'sale_name': CD(SaleNomenclature),
        'allow_edit': CD(bool),
        'denomination_id': CD(int)
    }

    def __init__(
        self,
        public_id: str,
        created: ProcuretTime,
        supplier: EntityHeadline,
        invitee_email: Optional[str],
        invitee_phone_digits: Optional[str],
        invitee_human_id: Optional[int],
        invoice_amount: Decimal,
        invoice_identifier: str,
        opens: List[InstalmentLinkOpen],
        sale_name: SaleNomenclature,
        allow_edit: bool,
        denomination_id: int
    ) -> None:

        self._supplier = supplier
        self._created = created
        self._public_id = public_id
        self._invitee_email = invitee_email
        self._invitee_phone_digits = invitee_phone_digits
        self._invitee_human_id = invitee_human_id
        self._invoice_amount = invoice_amount
        self._invoice_identifier = invoice_identifier
        self._opens = opens
        self._sale_name = sale_name
        self._allow_edit = allow_edit
        self._denomination_id = denomination_id

        return

    public_id = property(lambda s: s._public_id)
    created = property(lambda s: s._created)
    supplier = property(lambda s: s._supplier)
    invitee_email = property(lambda s: s._invitee_email)
    invitee_phone_digits = property(lambda s: s._invitee_phone_digits)
    invitee_human_id = property(lambda s: s._invitee_human_id)
    invoice_amount = property(lambda s: s._invoice_amount)
    invoice_identifier = property(lambda s: s._invoice_identifier)
    opens = property(lambda s: s._opens)
    sale_name = property(lambda s: s._sale_name)
    allow_edit = property(lambda s: s._allow_edit)
    denomination_id = property(lambda s: s._denomination_id)

    currency = property(lambda s: Currency.assertively_with_id(
        indexid=s._denomination_id
    ))

    invoice_amount_pretty = property(lambda s: '{:,}'.format(s.invoice_amount))
    has_been_opened = property(lambda s: len(s._opens) > 0)
    open_count = property(lambda s: len(s._opens))

    url = property(lambda s: s._LINK_TEMPLATE.format(public_id=s._public_id))

    amount: Amount = property(lambda s: Amount(
        magnitude=s._invoice_amount,
        denomination=Currency.assertively_with_id(s._denomination_id)
    ))

    @classmethod
    def create(
        cls: Type[Self],
        supplier: Union[int, EntityHeadline],
        invoice_amount: Decimal,
        invitee_email: str,
        invoice_identifier: str,
        communication: CommunicationOption,
        session: Session
    ) -> Self:

        def infer_supplier_id(x: Union[int, EntityHeadline]) -> int:
            if isinstance(x, int):
                return x
            if isinstance(x, EntityHeadline):
                return x.entity_id
            raise ProcuretTypeError(('int', 'EntityHeadline'), x, 'supplier')

        def infer_communication(y: CommunicationOption) -> bool:
            if not isinstance(communication, CommunicationOption):
                raise ProcuretTypeError(
                    'CommunicationOption',
                    y,
                    'communcation'
                )
            if communication == CommunicationOption.DO_NOT_CONTACT_CUSTOMER:
                return False
            if communication == CommunicationOption.EMAIL_CUSTOMER:
                return True
            raise NotImplementedError

        if not isinstance(invoice_amount, Decimal):
            raise ProcuretTypeError(
                'Decimal',
                invoice_amount,
                'invoice_amount'
            )

        if not isinstance(invitee_email, str):
            raise ProcuretTypeError('str', invitee_email, 'invitee_email')

        if not isinstance(invoice_identifier, str):
            raise ProcuretTypeError(
                'str',
                invoice_identifier,
                'invoice_identifier'
            )

        data = {
            'supplier_id': infer_supplier_id(supplier),
            'invoice_amount': str(invoice_amount),
            'invitee_email': invitee_email,
            'invoice_identifier': invoice_identifier,
            'communicate': infer_communication(communication)
        }

        result = ApiRequest.make(
            path=cls.path,
            method=HTTPMethod.POST,
            data=data,
            session=session,
            query_parameters=None
        )

        if result is None:
            raise InconsistentState

        return cls.decode(result)

    @classmethod
    def retrieve(
        cls: Type[Self],
        public_id: str,
        session: Session
    ) -> Optional[Self]:

        if not isinstance(public_id, str):
            raise TypeError('public_id must be a string')

        result = cls.retrieve_many(
            public_id=public_id,
            session=session
        )

        if len(result) < 1:
            return None

        return result[0]

    @classmethod
    def retrieve_many(
        cls: Type[Self],
        session: Session,
        public_id: Optional[str] = None,
        limit: int = 20,
        offset: int = 0,
        order: Order = Order.ASCENDING,
        order_by: OrderBy = OrderBy.CREATED,
        opened: Optional[bool] = None
    ) -> List[Self]:

        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')

        if not isinstance(offset, int):
            raise TypeError('offset must be an integer')

        if not isinstance(order, Order):
            raise TypeError('order must be of type Order')

        if not isinstance(order_by, OrderBy):
            raise TypeError('order must be of type InstalmentLink.OrderBy')

        if not isinstance(session, Session):
            raise TypeError('session must be of type Session')

        parameters = [
            QueryParameter('limit', limit),
            QueryParameter('offset', offset),
            QueryParameter('order', order.value),
            QueryParameter('order_by', order_by.value),
        ]

        if opened is not None:
            if not isinstance(opened, bool):
                raise TypeError('If supplied, opened must be a bool')
            parameters.append(QueryParameter('opened', opened))

        if public_id is not None:
            if not isinstance(public_id, str):
                raise TypeError('If supplied, public_id must be str')
            parameters.append(QueryParameter('public_id', public_id))

        result = ApiRequest.make(
            path=cls.list_path,
            method=HTTPMethod.GET,
            data=None,
            session=session,
            query_parameters=QueryParameters(parameters)
        )

        links = cls.optionally_decode_many(result, default_to_empty_list=True)
        assert links is not None
        return links



"""
Procuret Python
Currency Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.money.unit_of_account import UnitOfAccount
from typing import TypeVar, Type, Optional, Any
from decimal import Decimal

Self = TypeVar('Self', bound='Currency')


class Currency(Codable, UnitOfAccount):

    coding_map = {
        'indexid': CD(int),
        'name': CD(str),
        'iso_4217': CD(str),
        'exponent': CD(int),
        'symbol': CD(str)
    }

    def __init__(
        self,
        indexid: int,
        name: str,
        iso_4217: str,
        exponent: int,
        symbol: str
    ) -> None:

        self._indexid = indexid
        self._name = name
        self._iso_4217 = iso_4217
        self._exponent = exponent
        self._symbol = symbol

        return

    indexid = property(lambda s: s._indexid)
    exponent = property(lambda s: s._exponent)
    iso_4217 = property(lambda s: s._iso_4217)

    def to_pretty_string(self, amount: Decimal) -> str:
        amount = self.round_to_maximum_precision(amount)
        return '{0:,}'.format(amount)

    def to_symbolised_pretty_string(self, amount: Decimal) -> str:
        return self._symbol + self.to_pretty_string(amount)

    def encode(self) -> int:
        return self._indexid

    @classmethod
    def decode(cls: Type[Self], data: Any) -> Self:
        if isinstance(data, int):
            currency = cls.with_id(data)
            if currency is None:
                raise RuntimeError('Unknown currency ' + str(data))
            return currency
        return super().decode(data)

    @classmethod
    def with_iso4217(cls: Type[Self], iso_4217: str) -> Optional[Self]:
        iso_4217 = iso_4217.upper()
        if iso_4217 in Constants.ENUMERATIONS:
            return Constants.ENUMERATIONS[iso_4217]
        return None

    @classmethod
    def with_id(cls: Type[Self], indexid: int) -> Optional[Self]:
        for key in Constants.ENUMERATIONS:
            if Constants.ENUMERATIONS[key]._indexid == indexid:
                return Constants.ENUMERATIONS[key]
        return None

    @classmethod
    def assertively_with_id(cls: Type[Self], indexid: int) -> Self:
        currency = cls.with_id(indexid)
        if currency is None:
            raise RuntimeError('No currency found w/ id ' + str(indexid))
        return currency

    def __eq__(self, other):
        if other._iso_4217 == self._iso_4217:
            return True
        return False

    def __reduce__(self):
        return (Currency, (
            self._indexid,
            self._name,
            self._iso_4217,
            self._exponent,
            self._symbol
        ))


class Constants:

    AUD = Currency(
        indexid=1,
        name='Australian Dollar',
        iso_4217='AUD',
        exponent=2,
        symbol='$'
    )

    NZD = Currency(
        indexid=2,
        name='New Zealand Dollar',
        iso_4217='NZD',
        exponent=2,
        symbol='$'
    )

    ENUMERATIONS = {
        AUD.iso_4217: AUD,
        NZD.iso_4217: NZD
    }



"""
Procuret Python
Phone Number Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD


class PhoneNumber(Codable):

    coding_map = {
        'indexid': CD(int),
        'digits': CD(str),
        'confirmed': CD(bool, optional=True),
        'confirmation_required': CD(bool, optional=True),
        'confirmation_request_sent': CD(bool, optional=True)
    }

    def __init__(
        self,
        indexid: int,
        digits: str,
        confirmed: bool,
        confirmation_required: bool,
        confirmation_request_sent: bool
    ) -> None:

        self._indexid = indexid
        self._digits = digits
        self._confirmed = confirmed
        self._confirmation_required = confirmation_required
        self._confirmation_request_sent = confirmation_request_sent

        return

    indexid = property(lambda s: s._indexid)
    digits = property(lambda s: s._digits)
    confirmed = property(lambda s: s._confirmed)
    confirmation_required = property(lambda s: s._confirmation_required)
    confirmation_request_sent = property(
        lambda s: s._confirmation_request_sent
    )



"""
Procuret Python
Exercise Instalment Link Test
author: hugh@blinkybeach.com
"""
from procuret.integrate.xero import XeroEntityMap
from procuret.tests.test_result import Success, TestResult
from procuret.tests.variants.with_supplier import TestWithSupplier
from procuret.session import Perspective


class RetrieveXeroEntityMap(TestWithSupplier):

    NAME = 'Retrieve a Xero EntityMap'

    test_perspective = Perspective.SUPPLIER

    def execute(self) -> TestResult:

        entity_map = XeroEntityMap.retrieve(
            entity_id=self.supplier_id,
            session=self.session
        )

        # For the purposes of this test, an absent EntityMap with no errors
        # is an adequate outcome.
        assert entity_map is None or isinstance(entity_map, XeroEntityMap)

        return Success()



"""
Procuret Python
Human Identity Document
author: hugh@blinkybeach.com
"""
from procuret.human_identity_document.type import HumanIdentityDocumentType
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.time.time import ProcuretTime
from procuret.human.headline import HumanHeadline
from procuret.human_identity_document.state import HumanIdentityDocumentState
from procuret.data.disposition import Disposition


class HumanIdentityDocument(Codable):

    coding_map = {
        'public_id': CD(str),
        'created': CD(ProcuretTime),
        'document_type': CD(HumanIdentityDocumentType),
        'document_identifier': CD(str),
        'human': CD(HumanHeadline),
        'state': CD(HumanIdentityDocumentState),
        'active': CD(bool),
        'disposition': CD(Disposition)
    }

    def __init__(
        self,
        public_id: str,
        created: ProcuretTime,
        document_type: HumanIdentityDocumentType,
        document_identifier: str,
        human: HumanHeadline,
        state: HumanIdentityDocumentState,
        active: bool,
        disposition: Disposition
    ) -> None:

        self._public_id = public_id
        self._created = created
        self._document_type = document_type
        self._document_identifier = document_identifier
        self._human = human
        self._state = state
        self._active = active
        self._disposition = disposition

        return
    
    public_id = property(lambda s: s._public_id)
    created = property(lambda s: s._created)
    document_type = property(lambda s: s._document_type)
    document_identifier = property(lambda s: s._document_identifier)
    human = property(lambda s: s._human)
    state = property(lambda s: s._state)
    active = property(lambda s: s._active)
    disposition = property(lambda s: s._disposition)



"""
Procuret Python
Country Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD


class Country(Codable):

    coding_map = {
        'country_id': CD(int),
        'name': CD(str),
        'iso_3166_a2': CD(str),
        'iso_3166_a3': CD(str)
    }

    def __init__(
        self,
        country_id: int,
        name: CD(str),
        iso_3166_a2: CD(str),
        iso_3166_a3: CD(str)
    ) -> None:

        self._country_id = country_id
        self._name = name
        self._iso_3166_a2 = iso_3166_a2
        self._iso_3166_a3 = iso_3166_a3

        return

    country_id = property(lambda s: s._country_id)
    name = property(lambda s: s._name)
    iso_3166_a2 = property(lambda s: s._iso_3166_a2)
    iso_3166_a3 = property(lambda s: s._iso_3166_a3)



"""
Procuret Python
Extract Argument Function
author: hugh@blinkybeach.com
"""
import sys
from typing import Optional, Dict, List, TypeVar, Type
from typing import Any

T = TypeVar('T', bound='CommandLine')


class CommandLine:

    def __init__(
        self,
        arguments: List[str]
    ) -> None:

        data: Dict[str, Optional[str]] = dict()
        index = 0
        while index < len(arguments):
            value: Optional[str] = None
            argument = arguments[index]
            if argument[0] == '-':
                n = index + 1
                if (n) < len(arguments) and arguments[n][0] != '-':
                    value = arguments[n]
                    index += 1
                else:
                    value = None
            data[argument] = value
            index += 1
            continue

        self._data = data
        return

    def contains_flag(self, flag: str) -> bool:
        if flag in self._data.keys():
            return True
        return False

    def get(
        self,
        key: str,
        of_type: Optional[Type] = None,
        type_name: Optional[str] = None
    ) -> Optional[Any]:

        if key not in self._data.keys():
            return None

        value = self._data[key]

        if of_type is not None and not isinstance(value, of_type):
            try:
                value = of_type(value)
            except Exception:
                raise ValueError('Value for parameter {k} must be a string tha\
t may be cast to {t}'.format(k=key, t=(type_name or str(of_type))))

        return value

    def require(
        self,
        key: str,
        of_type: Optional[Type] = None,
        type_name: Optional[str] = None
    ) -> Any:

        value = self.get(key, of_type, type_name=type_name)
        if value is None:
            raise RuntimeError('Missing command line parmeter "{p}"'.format(
                p=key
            ))

        return value

    @classmethod
    def load(cls: Type[T]) -> T:
        return cls(sys.argv[1:])



"""
Procuret Python
Human Identity Document Type Module
author: hugh@blinkybeach.com
"""
from enum import Enum


class HumanIdentityDocumentType(Enum):

    AUS_DRIVER_LICENSE = 1
    NZ_DRIVER_LICENSE = 2



"""
Procuret API Python
Sale Nomenclature Module
author: hugh@blinkybeach.com
"""
from enum import IntEnum


class SaleNomenclature(IntEnum):

    INVOICE = 1
    ORDER = 2
    SHOPIFY_ORDER = 3
    REFERENCE = 4



"""
Procuret Python
HumanIdentity Module
author: hugh@blinkybeach.com
"""
from typing import Optional
from procuret.time.time import ProcuretTime
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.address.address import Address


class HumanIdentity(Codable):

    coding_map = {
        'indexid': CD(int),
        'created': CD(ProcuretTime),
        'modified': CD(ProcuretTime),
        'date_of_birth': CD(str, optional=True),
        'address': CD(Address, optional=True)
    }

    def __init__(
        self,
        indexid: int,
        created: ProcuretTime,
        modified: ProcuretTime,
        date_of_birth: Optional[str],
        address: Optional[Address]
    ) -> None:

        self._indexid = indexid
        self._created = created
        self._modified = modified
        self._date_of_birth = date_of_birth
        self._address = address

    indexid = property(lambda s: s._indexid)
    created = property(lambda s: s._created)
    modified = property(lambda s: s._modified)
    date_of_birth = property(lambda s: s._date_of_birth)
    address = property(lambda s: s._address)



"""
Procuret API
Procuret Time Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable
from datetime import timedelta
from datetime import datetime
from typing import TypeVar, Type, Any
from procuret.time.tz_utc import UTC

T = TypeVar('T', bound='ProcuretTime')


class ProcuretTime(datetime, Codable):
    _DB_FORMAT_STRING = '%Y-%m-%d %H:%M:%S.%f'

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:

        assert isinstance(data, str)
        if '+' in data:
            parsable = data.split('+')[0]
        else:
            if '-' in data.split(':')[2]:
                parsable = data[:-6]
            else:
                parsable = data
        parsable = parsable.replace('T', ' ')
        parsable = parsable.replace('_', ' ')
        time = cls.strptime(parsable, cls._DB_FORMAT_STRING)
        decoded = cls(
            year=time.year,
            month=time.month,
            day=time.day,
            hour=time.hour,
            minute=time.minute,
            second=time.second,
            microsecond=time.microsecond,
            tzinfo=UTC
        )

        return decoded

    def encode(self) -> str:
        return self.strftime(self._DB_FORMAT_STRING)

    @classmethod
    def utcnow(cls: Type[T]) -> T:
        time = datetime.utcnow()
        return cls._from_datetime(time)

    @classmethod
    def from_unix_timestamp(cls: Type[T], timestamp: int) -> T:
        time = datetime.fromtimestamp(timestamp)
        return cls._from_datetime(time)

    @classmethod
    def in_seconds_from_now(cls: Type[T], seconds: int) -> T:
        time = datetime.utcnow() + timedelta(seconds=seconds)
        return cls._from_datetime(time)

    @classmethod
    def in_minutes_from_now(cls: Type[T], minutes: int) -> T:
        time = datetime.utcnow() + timedelta(minutes=minutes)
        return cls._from_datetime(time)

    @classmethod
    def in_days_from_now(cls: Type[T], days: int) -> T:
        time = datetime.utcnow() + timedelta(days=days)
        return cls._from_datetime(time)

    @classmethod
    def _from_datetime(cls: Type[T], time: datetime) -> T:
        return cls(
            year=time.year,
            month=time.month,
            day=time.day,
            hour=time.hour,
            minute=time.minute,
            second=time.second,
            microsecond=time.microsecond,
            tzinfo=UTC
        )



"""
Procuret Python
QueryParameter Module
author: hugh@blinkybeach.com
"""
from typing import Any, TypeVar, Type, List, Optional

Self = TypeVar('Self', bound='QueryParameter')


class QueryParameter:
    """A single URL parameter, e.g. beep=boop"""
    def __init__(
        self,
        key: str,
        value: Any
    ) -> None:

        assert isinstance(key, str)
        str(value)  # provoke error early
        self._key = key
        self._value = value

        self._url_representation = self._represent(value)

        return

    key = property(lambda s: s._key)

    def __str__(self) -> str:
        return self._key + '=' + self._url_representation

    @staticmethod
    def remove_targets_with(
        key: str,
        targets: List[Self]
    ) -> List[Self]:

        retained_targets: List[QueryParameter] = list()
        for target in targets:
            if target._key == key:
                continue
            retained_targets.append(target)
            continue

        return targets

    @classmethod
    def optionally(
        Self: Type[Self],
        key: str,
        value: Optional[Any]   
    ) -> Optional[Self]:

        if value is None:
            return None
        
        return Self(key, value)


    @staticmethod
    def _represent(value: Any) -> str:

        if isinstance(value, str):
            return value

        if isinstance(value, bool):
            if value is True:
                return 'true'
            return 'false'

        return str(value)



"""
Procuret Python
Exercise Term Rate Group Test
author: hugh@blinkybeach.com
"""
from procuret.tests.variants.with_supplier import TestWithSupplier
from procuret.tests.test_result import Success, TestResult
from procuret.session import Perspective
from procuret.ancillary.command_line import CommandLine
from procuret.term_rate.group import TermRateGroup


class ExerciseTermRateGroup(TestWithSupplier):

    NAME = 'Retrieve a TermRateGroup'

    test_perspective = Perspective.SUPPLIER

    def execute(self) -> TestResult:

        cl = CommandLine.load()

        group_id = cl.require(
            key='--term-rate-group',
            of_type=str,
            type_name='string'
        )

        group = TermRateGroup.retrieve(
            public_id=group_id,
            session=self.session
        )

        assert isinstance(group, TermRateGroup)

        return Success()



"""
Procuret Python
Session Perspective Module
author: hugh@blinkybeach.com
"""
from enum import Enum


class Perspective(Enum):
    SUPPLIER = 1
    BUSINESS = 2
    ADMINISTRATOR = 3
    SALES = 4
    TREASURY = 5
    INVESTOR = 6
    CUSTOMER_SUPPORT = 7



"""
Procuret Python
Payment Method Headline Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD


class PaymentMethodHeadline(Codable):

    coding_map = {
        'public_id': CD(str),
        'description': CD(str)
    }

    def __init__(
        self,
        public_id: str,
        description: str
    ) -> None:

        self._public_id = public_id
        self._description = description

        return

    public_id = property(lambda s: s._public_id)
    description = property(lambda s: s._description)



"""
Procuret Python
Disposition Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable
from typing import TypeVar, Type, Any, Dict

T = TypeVar('T', bound='Disposition')


class Disposition(Codable):

    def __init__(
        self,
        sequence: int,
        count: int,
        limit: int,
        offset: int
    ) -> None:

        self._sequence = sequence
        self._count = count
        self._limit = limit
        self._offset = offset

        return

    sequence = property(lambda s: s._sequence)
    count = property(lambda s: s._count)
    limit = property(lambda s: s._limit)
    offset = property(lambda s: s._offset)

    def encode(self) -> Dict[str, int]:
        return {
            'sequence': self._sequence,
            'count': self._count,
            'limit': self._limit,
            'offset': self._offset
        }

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            sequence=data['sequence'],
            count=data['count'],
            limit=data['limit'],
            offset=data['offset']
        )



from procuret.errors.api_error import ProcuretApiError
from procuret.errors.error import ProcuretError
from procuret.errors.inconsistent import InconsistentState
from procuret.errors.type_error import ProcuretTypeError



"""
Procuret Python
Decodable Module
author: hugh@blinkybeach.com
"""
from json import loads
from typing import Any, Optional, TypeVar, Type, List

T = TypeVar('T', bound='Decodable')


class Decodable:
    """Abstract protocol defining an interface for decodable classes"""

    @classmethod
    def decode(self, data: Any) -> T:
        """Return a JSON-serialisable form of the object"""
        raise NotImplementedError

    @classmethod
    def optionally_decode(cls: Type[T], data: Optional[Any]) -> Optional[T]:
        """Optionally return a decoded object from serialised data"""
        if data is None:
            return None
        return cls.decode(data)

    @classmethod
    def deserialise(cls: Type[T], serial: str) -> T:
        """Return a JSON string representation of the object"""
        return cls.decode(loads(serial))

    @classmethod
    def optionally_deserialise(
        cls: Type[T],
        serial: Optional[str]
    ) -> Optional[T]:
        if serial is None:
            return None
        return cls.deserialise(serial)

    @classmethod
    def decode_many(cls: Type[T], data: Any) -> List[T]:
        """Return list of decoded instances of an object"""
        return [cls.decode(d) for d in data]

    @classmethod
    def optionally_decode_many(
        cls: Type[T],
        data: Optional[Any],
        default_to_empty_list: bool = False
    ) -> Optional[List[T]]:
        """Optionally return a list of decoded objects"""
        if data is None and default_to_empty_list is True:
            return list()
        if data is None:
            return None
        return cls.decode_many(data)



"""
Procuret Python
Abstract Session Module
author: hugh@blinkybeach.com
"""
from procuret.ancillary.session_lifecycle import Lifecycle
from procuret.ancillary.session_perspective import Perspective
from procuret.data.codable import Codable
from typing import Optional


class AbstractSession(Codable):

    session_id: int
    session_key: str
    api_key: str
    lifecycle: Lifecycle
    perspective: Perspective

    acts_for_another_agent: bool
    on_behalf_of: Optional[int]



"""
Procuret Python
Api Error Module
author: hugh@blinkybeach.com
"""
from procuret.errors.error import ProcuretError


class ProcuretApiError(RuntimeError, ProcuretError):

    def __init__(
        self,
        http_code: int
    ) -> None:
        super().__init__('Procuret API responded to your request with an error\
code - {c}. There may be a bug in the Procuret Python library, or Procuret API\
 may be experiencing a temporary disruption. If this happens repeatedly, pleas\
e contact us at support@procuret.com'.format(c=str(http_code)))



"""
Procuret Python
Amount Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.money.currency import Currency, Constants as Currencies
from decimal import Decimal
from typing import Type, TypeVar, List, Any

Self = TypeVar('Self', bound='Amount')


class Amount(Codable):

    coding_map = {
        'magnitude': CD(Decimal),
        'denomination': CD(Currency)
    }

    def __init__(
        self,
        magnitude: Decimal,
        denomination: Currency
    ) -> None:

        self._magnitude = magnitude
        self._denomination = denomination

        return

    magnitude = property(lambda s: s._magnitude)
    denomination = property(lambda s: s._denomination)

    string_magnitude = property(str(lambda s: s._magnitude))
    pretty_magnitude = property(
        lambda s: '{:,}'.format(s._magnitude)
    )

    as_integer_smallest_unit= property(
        lambda s: s._denomination.to_integer_smallest_unit(s._magnitude)
    )

    as_pretty_string = property(
        lambda s: s._denomination.to_pretty_string(s._magnitude)
    )

    as_symbolised_pretty_string = property(
        lambda s: s._denomination.to_symbolised_pretty_string(s._magnitude)
    )

    as_denominated_pretty_string = property(
        lambda s: '{d} {a}'.format(
            d=s.denomination.iso_4217.upper(),
            a=s.as_symbolised_pretty_string
        )
    )

    def is_denominated_in(self, currency: Currency) -> bool:
        if self._denomination.indexid == currency.indexid:
            return True
        return False

    @classmethod
    def decode(Self: Type[Self], data: Any) -> Self:

        if isinstance(data['denomination'], int):
            return Self(
                magnitude=Decimal(data['magnitude']),
                denomination=Currency.with_id(data['denomination'])
            )

        return super().decode(data)

    @classmethod
    def denominations_are_homogenous(
        Self: Type[Self],
        amounts: List[Self]
    ) -> bool:
        unique = set(amounts)
        if len(unique) > 1:
            return False
        return True

    def _affirm_comparability(self, other: Self) -> None:
        if not isinstance(other, Amount):
            raise TypeError('comparison candidate must be of type Amount')
        if other.denomination != self._denomination:
            raise ValueError('comparison candiates must share denomination')
        return None

    def __lt__(self, other: Self) -> bool:
        self._affirm_comparability(other)
        return (self._magnitude < other._magnitude)

    def __le__(self, other: Self) -> bool:
        self._affirm_comparability(other)
        return (self._magnitude <= other._magnitude)

    def __eq__(self, other: Self) -> bool:
        self._affirm_comparability(other)
        return self._magnitude == other._magnitude

    def __ne__(self, other: Self) -> bool:
        self._affirm_comparability(other)
        return self._magnitude != other.magnitude

    def __gt__(self, other: Self) -> bool:
        self._affirm_comparability(other)
        return (self._magnitude > other._magnitude)

    def __ge__(self, other: Self) -> bool:
        self._affirm_comparability(other)
        return (self._magnitude >= other._magnitude)

    def __add__(self, other: Self) -> Self:
        self._affirm_comparability(other)
        return Amount(
            magnitude=self._magnitude + other._magnitude,
            denomination=self._denomination
        )

    def __radd__(self, other: Any) -> Self:
        if other != 0:
            raise RuntimeError('Cannot add non-zero un-denominated value to \
Amount')
        return self

    def __sub__(self, other: Self) -> Self:
        self._affirm_comparability(other)
        return Amount(
            magnitude=self._magnitude - other._magnitude,
            denomination=self._denomination
        )

    def __mul__(self, other: Any) -> Self:
        if (isinstance(other, Amount)):
            self._affirm_comparability(other)
            return Amount(
                magnitude=self._magnitude * other._magnitude,
                denomination=self._denomination
            )
        return Amount(
            magnitude=self._magnitude * Decimal(other),
            denomination=self._denomination
        )

    def __rmul__(self, other: Any) -> Self:
        return self.__mul__(other)

    def __truediv__(self, other: Any) -> Self:
        if (isinstance(other, Amount)):
            self._affirm_comparability(other)
            return Amount(
                magnitude=self._magnitude / other._magnitude,
                denomination=self._denomination
            )
        return Amount(
            magnitude=self._magnitude / Decimal(other),
            denomination=self._denomination
        )

    def __repr__(self) -> str:
        return 'Amount: {a}'.format(a=self.as_denominated_pretty_string)

    def __str__(self) -> str:
        return self.as_denominated_pretty_string

    def __round__(self, exponent: int) -> Self:
        return Amount(
            magnitude=round(self.magnitude, exponent),
            denomination=self.denomination
        )

    @staticmethod
    def make_test_amount() -> Self:
        return Amount(Decimal(420), Currencies.AUD)



from procuret.session import Session, Lifecycle, Perspective, AbstractSession
from procuret.ancillary.communication_option import CommunicationOption
from procuret.instalment_link import InstalmentLink, InstalmentLinkOpen
from procuret.instalment_link import InstalmentLinkOrderBy
from procuret.data.order import Order
from procuret.human.headline import HumanHeadline
from procuret.ancillary.entity_headline import EntityHeadline
from procuret.ancillary.sale_nomenclature import SaleNomenclature
from procuret import errors
from procuret.version import VERSION
from procuret.term_rate.term_rate import TermRate
from procuret.term_rate.group import TermRateGroup
from procuret.integrate.xero import XeroOrganisation, XeroEntityMap
from procuret.money.amount import Amount
from procuret.money.currency import Currency, Constants as Currencies
from procuret.data.codable import Codable, CodingDefinition
from procuret.global_brand.brand import GlobalBrand
from procuret.global_brand.selection import GlobalBrandSelection
from procuret.payment_series.series import PaymentSeries
from procuret.customer_payment import CustomerPayment
from procuret.payment_series.payment_mechanism import PaymentMechanism
from procuret.human.human import Human



"""
Procuret Python
Payment Mechanism Module
author: inkalchemi@gmail.com
"""
from enum import IntEnum
from typing import TypeVar

Self = TypeVar('Self', bound='PaymentMechanism')


class PaymentMechanism(IntEnum):

    ON_PLATFORM = 1
    OFF_PLATFORM = 2
    STRIPE_SUBSCRIPTION = 3
    UNKNOWN = 4


    @staticmethod
    def name_for(mechanism: Self) -> str:
        return {
            PaymentMechanism.ON_PLATFORM: 'On Platform',
            PaymentMechanism.OFF_PLATFORM: 'Off Platform',
            PaymentMechanism.UNKNOWN: 'Unknown',
            PaymentMechanism.STRIPE_SUBSCRIPTION: 'Stripe Subscription'
        }[mechanism]



"""
Procuret Python
Type Error Module
author: hugh@blinkybeach.com
"""
from procuret.errors.error import ProcuretError
from typing import Union, Tuple, Any


class ProcuretTypeError(TypeError, ProcuretError):

    _PLURAL_TEMPLATE = '{v} must be an instance of one of types {t1} or {t2}, \
not {s}'

    _SINGULAR_TEMPLATE = '{v} must be an instance of type {t}, not {s}'

    def __init__(
        self,
        valid_types: Union[Tuple[str], str],
        supplied_variable: Any,
        variable_name: str
    ) -> None:

        if isinstance(valid_types, tuple):
            return super().__init__(self._PLURAL_TEMPLATE.format(
                v=variable_name,
                t1=', '.join(str(valid_types[:-1])),
                t2=valid_types[-1],
                s=str(type(supplied_variable))
            ))

        return super().__init__(self._SINGULAR_TEMPLATE.format(
            v=variable_name,
            t=valid_types,
            s=str(type(supplied_variable))
        ))



"""
Procuret Python
Customer Payment Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD 
from procuret.time.time import ProcuretTime
from decimal import Decimal
from procuret.human.headline import HumanHeadline
from procuret.data.disposition import Disposition
from procuret.payment_method.headline import PaymentMethodHeadline
from procuret.money.currency import Currency
from typing import Optional


class CustomerPayment(Codable):

    coding_map = {
        'public_id': CD(str),
        'executor': CD(HumanHeadline),
        'created': CD(ProcuretTime),
        'executed': CD(ProcuretTime),
        'magnitude': CD(Decimal),
        'disposition': CD(Disposition),
        'active': CD(bool),
        'payment_method': CD(PaymentMethodHeadline, optional=True),
        'commitment_id': CD(str),
        'denomination': CD(Currency)
    }

    def __init__(
        self,
        public_id: str,
        executor: HumanHeadline,
        created: ProcuretTime,
        executed: ProcuretTime,
        magnitude: Decimal,
        disposition: Disposition,
        active: bool,
        payment_method: Optional[PaymentMethodHeadline],
        commitment_id: str,
        denomination: Currency
    ) -> None:

        self._public_id = public_id
        self._executor = executor
        self._created = created
        self._executed = executed
        self._disposition = disposition
        self._magnitude = magnitude
        self._active = active
        self._payment_method = payment_method
        self._commitment_id = commitment_id
        self._denomination = denomination

        return

    public_id: str = property(lambda s: s._public_id)
    executor: HumanHeadline = property(lambda s: s._executor)
    created: ProcuretTime = property(lambda s: s._created)
    executed: ProcuretTime = property(lambda s: s._executed)
    magnitude: Decimal = property(lambda s: s._magnitude)
    disposition: Disposition = property(lambda s: s._disposition)
    active: bool = property(lambda s: s._active)
    payment_method: Optional[PaymentMethodHeadline] = property(
        lambda s: s._payment_method
    )
    commitment_id: str = property(lambda s: s._commitment_id)
    denomination: Currency = property(lambda s: s._denomination)



"""
Procuret Python
TimeZoneUTC Module
author: hugh@blinkybeach.com
"""
import datetime


class TimeZoneUTC(datetime.tzinfo):
    """
    UTC timezone.
    """
    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return 'UTC'

    def dst(self, dt):
        return datetime.timedelta(0)


UTC = TimeZoneUTC()



"""
Procuret Python
Procuret Date Module
author: hugh@blinkybeach.com
"""
from datetime import datetime
from typing import TypeVar, Type, Any
from procuret.data.codable import Codable, CodingDefinition as CD

T = TypeVar('T', bound='ProcuretDate')


class ProcuretDate(datetime, Codable):
    _DB_FORMAT_STRING = "%Y-%m-%d"
    _REQUEST_FORMAT_STRING = "%Y-%m-%d"
    _REQUEST_FORMAT_STRING_B = "%Y/%m/%d"
    _REQUEST_FORMAT_STRING_C = "%d/%m/%Y"
    _REQUEST_FORMAT_STRING_D = "%d-%m-%Y"

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:

        assert isinstance(data, str)
        data = data.split('T')[0]
        date = cls.strptime(data, cls._DB_FORMAT_STRING)
        return cls(
            year=date.year,
            month=date.month,
            day=date.day
        )

    def encode(self) -> str:
        return self.strftime(self._REQUEST_FORMAT_STRING)

    @classmethod
    def today(cls: Type[T]) -> T:
        return cls.create_now()

    @classmethod
    def create_now(cls: Type[T]) -> T:
        """Return the current date, at UTC"""
        date = datetime.utcnow()
        return cls(
            year=date.year,
            month=date.month,
            day=date.day
        )




"""
Procuret Python
Inconsistent State Error
author: hugh@blinkybeach.com
"""
from procuret.errors.error import ProcuretError


class InconsistentState(RuntimeError, ProcuretError):

    def __init__(self) -> None:
        super().__init__('Procuret API responded in a manner inconsistent wit\
h Procuret Python\'s expectations. This may indicate that there is a bug in t\
he Procuret Python library, or that Procuret API has responded in a manner in\
consistent with its documentation. Please contact us at support@procuret.com')



"""
Procuret Python
Human Headine Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD


class HumanHeadline(Codable):

    coding_map = {
        'agent_id': CD(int),
        'full_name': CD(str)
    }

    def __init__(
        self,
        agent_id: int,
        full_name: str
    ) -> None:

        self._agent_id = agent_id
        self._full_name = full_name

        return

    full_name = property(lambda s: s._full_name)
    agent_id = property(lambda s: s._agent_id)



"""
Procuret Python
Payment Series Module
author: hugh@blinkybeach.com
"""
from enum import Enum
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.money.amount import Amount
from procuret.ancillary.entity_headline import EntityHeadline
from procuret.time.time import ProcuretTime
from procuret.payment_method.headline import PaymentMethodHeadline
from procuret.data.disposition import Disposition
from typing import TypeVar, Type, Optional, List
from procuret.data.order import Order
from procuret.session import Session
from procuret.http.api_request import ApiRequest, HTTPMethod, QueryParameters
from procuret.http.query_parameter import QueryParameter
from procuret.payment_series.payment_mechanism import PaymentMechanism

Self = TypeVar('Self', bound='PaymentSeries')


class OrderBy(Enum):
    CREATED = 'created'


class PaymentSeries(Codable):

    path = '/payment/series'

    coding_map = {
        'public_id': CD(str),
        'created': CD(ProcuretTime),
        'creating_agent': CD(int),
        'payment_method': CD(PaymentMethodHeadline),
        'customer': CD(EntityHeadline),
        'supplier': CD(EntityHeadline),
        'exchange_id': CD(str),
        'amount': CD(Amount),
        'sum_payments': CD(Amount),
        'total_payable': CD(Amount),
        'identifier': CD(str),
        'mechanism': CD(PaymentMechanism),
        'disposition': CD(Disposition)
    }

    OrderBy = OrderBy
  
    def __init__(
        self,
        public_id: str,
        created: ProcuretTime,
        creating_agent: str,
        payment_method: PaymentMethodHeadline,
        customer: EntityHeadline,
        supplier: EntityHeadline,
        exchange_id: str,
        amount: Amount,
        sum_payments: Amount,
        total_payable: Amount,
        identifier: str,
        mechanism: PaymentMechanism,
        disposition: Disposition
    ) -> None:

        self._public_id = public_id
        self._created = created
        self._creating_agent = creating_agent
        self._payment_method = payment_method
        self._customer = customer
        self._supplier = supplier
        self._exchange_id = exchange_id
        self._amount = amount
        self._sum_payments = sum_payments
        self._total_payable = total_payable
        self._identifier = identifier
        self._mechanism = mechanism
        self._disposition = disposition

        return

    public_id: str = property(lambda s: s._public_id)
    created: ProcuretTime = property(lambda s: s._created)
    creating_agent: int = property(lambda s: s._creating_agent)
    payment_method: PaymentMethodHeadline = property(
        lambda s: s._payment_method
    )
    customer: EntityHeadline = property(lambda s: s._customer)
    supplier: EntityHeadline = property(lambda s: s._supplier)
    exchange_id: str = property(lambda s: s._exchange_id)
    amount: Amount = property(lambda s: s._amount)
    sum_payments: Amount = property(lambda s: s._sum_payments)
    total_payable: Amount = property(lambda s: s._total_payable)
    identifier: str = property(lambda s: s._identifier)
    mechanism: PaymentMechanism = property(lambda s: s._mechanism)
    disposition: Disposition = property(lambda s: s._disposition)

    @classmethod
    def retrieve(
        Self: Type[Self],
        session: Session,
        public_id: str
    ) -> Optional[Self]:

        if not isinstance(public_id, str):
            raise TypeError('`public_id` must be of type `str`')

        result = Self.retrieve_many(
            session=session,
            public_id=public_id
        )

        if len(result) < 1:
            return None
        
        return result[0]

    @classmethod
    def retrieve_many(
        Self: Type[Self],
        session: Session,
        public_id: Optional[str] = None,
        text_fragment: Optional[str] = None,
        business_id: Optional[str] = None,
        method_id: Optional[str] = None,
        limit: int = 20,
        offset: int = 0,
        order: Order = Order.DESCENDING,
        order_by: OrderBy = OrderBy.CREATED
    ) -> List[Self]:

        QP = QueryParameter
        aq = [
            QP('limit', limit),
            QP('offset', offset),
            QP('order', order.value),
            QP('order_by', order_by.value),
            QP.optionally('any_fragment', text_fragment),
            QP.optionally('business_id', business_id),
            QP.optionally('method_id', method_id),
            QP.optionally('public_id', public_id)
        ]

        result = ApiRequest.make(
            path=Self.path + '/list',
            method=HTTPMethod.GET,
            data=None,
            session=session,
            query_parameters=QueryParameters([q for q in aq if q is not None])
        )

        return Self.decode_many(result) or []



"""
Procuret Python
Instalment Line Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.time.date import ProcuretDate
from decimal import Decimal
from typing import Optional
from procuret.money.currency import Currency
from procuret.customer_payment.payment import CustomerPayment


class InstalmentLine(Codable):

    coding_map = {
        'sequence': CD(int),
        'date': CD(ProcuretDate),
        'nominal_payment': CD(Decimal),
        'opening_balance': CD(Decimal),
        'interest_paid': CD(Decimal),
        'principal_paid': CD(Decimal),
        'principal_discount': CD(Decimal),
        'closing_balance': CD(Decimal),
        'commitment_public_id': CD(str),
        'payment': CD(CustomerPayment, optional=True),
        'denomination': CD(Currency)
    }

    def __init__(
        self,
        sequence: int,
        date: ProcuretDate,
        nominal_payment: Decimal,
        opening_balance: Decimal,
        interest_paid: Decimal,
        principal_paid: Decimal,
        principal_discount: Decimal,
        closing_balance: Decimal,
        commitment_public_id: str,
        payment: Optional[CustomerPayment],
        denomination: Currency
    ) -> None:

        self._sequence = sequence
        self._date = date
        self._nominal_payment = nominal_payment
        self._opening_balance = opening_balance
        self._interest_paid = interest_paid
        self._principal_paid = principal_paid
        self._princiapl_discount = principal_discount
        self._closing_balance = closing_balance
        self._commitment_public_id = commitment_public_id
        self._payment = payment
        self._denomination = denomination

        return

    sequence: int = property(lambda s: s._sequence)
    date: ProcuretDate = property(lambda s: s._date)
    nominal_payment: Decimal = property(lambda s: s._nominal_payment)
    opening_balance: Decimal = property(lambda s: s._opening_balance)
    interest_paid: Decimal = property(lambda s: s._interest_paid)
    principal_paid: Decimal = property(lambda s: s._principal_paid)
    principal_discount: Decimal = property(lambda s: s._principal_discount)
    closing_balance: Decimal = property(lambda s: s._closing_balance)
    commitment_public_id: Decimal = property(lambda s: s._commitment_public_id)
    payment: Optional[CustomerPayment] = property(lambda s: s._payment)
    denomination: Currency = property(lambda s: s._denomination)



"""
Procuret Python
Codable Module
author: hugh@procuret.com
"""
import sys
from typing import TypeVar, Type, Any, Union, Dict, Optional, List
from typing import Callable
from enum import Enum
from decimal import Decimal
from procuret.data.decodable import Decodable

T = TypeVar('T', bound='Codable')
K = TypeVar('K', bound='CodableType')

Object = Dict[str, K]
Array = List[K]
CodableType = Union[T, int, Object, Array, str, float, Enum, Decimal]

_QUANTUM_TYPES = (int, str, float, dict, list, bool)


def _recurse_bases(target: Type, class_definition: Type) -> bool:
    if target in class_definition.__bases__:
        return True
    if len(class_definition.__bases__) < 1:
        return False
    for base in class_definition.__bases__:
        result = _recurse_bases(target, base)
        if result is True:
            return result
        continue
    return False


class CodingDefinition:

    def __init__(
        self,
        codable_type: Type[CodableType],
        optional: bool = False,
        array: bool = False,
        default_value_generator: Optional[Callable[[], CodableType]] = None
    ) -> None:

        if (
                codable_type not in _QUANTUM_TYPES
                and not _recurse_bases(Enum, codable_type)
                and codable_type != Decimal
                and not _recurse_bases(Codable, codable_type)
        ):
            raise TypeError('CodingDefinition  requires that the `codable_type\
` parameter be one of either (int, str, float, bool, dict, list, Enum, Decimal\
) or be a custom type that inherits from Nozomi.Codable. The supplied codable_\
type {x} does not appear to meet these requirements.\
'.format(x=str(codable_type)))

        default_value: Optional[CodableType] = None
        if default_value_generator is not None:
            default_value = default_value_generator()

        self._codable_type = codable_type
        self._optional = optional
        self._array = array
        self._default_value = default_value

        return

    def decode(self, data: Any) -> Any:

        try:

            if data is None:
                if self._optional is True:
                    if self._default_value is not None:
                        return self._default_value
                    return None
                raise RuntimeError('Unexpectedly null data when decoding')

            if _recurse_bases(Enum, self._codable_type):
                if self._array is True:
                    return [self._codable_type(d) for d in data]
                return self._codable_type(data)

            if self._codable_type == Decimal:
                return Decimal(data)

            if not _recurse_bases(Codable, self._codable_type):
                return data

            if self._array is True and self._optional is True:
                return self._codable_type.optionally_decode_many(
                    data=data,
                    default_to_empty_list=True
                )
            if self._array is True:
                return self._codable_type.decode_many(data)
            if self._optional is True:
                return self._codable_type.optionally_decode(data)
            return self._codable_type.decode(data)
        except Exception:
            if '--debug' in sys.argv or '--test' in sys.argv:
                print('Failed to decode {t}:'.format(
                    t=str(self._codable_type)
                ))
                print(data)
            raise


class Codable(Decodable):

    coding_map: Dict[str, CodingDefinition] = NotImplemented

    def encode(self):

        def recurse(
            attribute: Optional[CodableType]
        ) -> Optional[CodableType]:

            if attribute is None:
                return None

            if type(attribute) in (int, str, float, bool):
                return attribute

            if isinstance(attribute, Enum):
                return recurse(attribute.value)

            if isinstance(attribute, Decimal):
                return str(attribute)

            if isinstance(attribute, dict):
                return {k: recurse(attribute[k]) for k in attribute}

            if isinstance(attribute, list):
                return [recurse(a) for a in attribute]

            if _recurse_bases(Codable, type(attribute)):
                return attribute.encode()

            raise RuntimeError('Unexpected type: ' + str(attribute))

        return {k: recurse(self.__dict__['_' + k]) for k in self.coding_map}

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        c = cls.coding_map
        return cls(**{k: c[k].decode(data[k]) for k in c})



"""
Procuret Python
Second Factor Code Module
author: hugh@blinkybeach.com
"""
from pydoc import plain
from procuret.ancillary.session_perspective import Perspective
from typing import Optional
from procuret.http.method import HTTPMethod
from procuret.http.api_request import ApiRequest


class SecondFactorCode:

    PATH = '/second-factor-code'

    @staticmethod
    def create_with_email(
        email: str,
        plaintext_secret: str,
        perspective: Optional[Perspective]
    ) -> None:

        return SecondFactorCode._create(
            email=email,
            agent_id=None,
            plaintext_secret=plaintext_secret,
            perspective=perspective
        )

    @staticmethod
    def create_with_agent_id(
        agent_id: int,
        plaintext_secret: str,
        perspective: Optional[Perspective]
    ) -> None:

        return SecondFactorCode._create(
            email=None,
            agent_id=agent_id,
            plaintext_secret=plaintext_secret,
            perspective=perspective
        )

    @staticmethod
    def _create(
        email: Optional[str],
        agent_id: Optional[int],
        plaintext_secret: str,
        perspective: Optional[Perspective]
    ) -> None:

        data = {
            'email': email,
            'plaintext_secret': plaintext_secret,
            'agent_id': agent_id,
            'perspective': (
                perspective.value if perspective is not None else None
            )
        }

        ApiRequest.make(
            path=SecondFactorCode.PATH,
            method=HTTPMethod.POST,
            data=data,
            session=None,
            query_parameters=None
        )

        return None



"""
Procuret Python
Global Brand Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD


class GlobalBrand(Codable):

    coding_map = {
        'indexid': CD(int),
        'name': CD(str),
        'logo_url_light': CD(str),
        'logo_url_dark': CD(str)
    }

    def __init__(
        self,
        indexid: int,
        name: str,
        logo_url_light: str,
        logo_url_dark: str
    ) -> None:

        self._indexid = indexid
        self._name = name
        self._logo_url_light = logo_url_light
        self._logo_url_dark = logo_url_dark

        return

    indexid = property(lambda s: s._indexid)
    name = property(lambda s: s._name)
    logo_url_light = property(lambda s: s._logo_url_light)
    logo_url_dark = property(lambda s: s._logo_url_dark)



"""
Procuret Python
Region Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD


class Region(Codable):

    coding_map = {
        'indexid': CD(int),
        'name': CD(str),
        'abbreviation': CD(str),
        'country_id': CD(int)
    }

    def __init__(
        self,
        indexid: int,
        name: str,
        abbreviation: str,
        country_id: int
    ) -> None:

        self._indexid = indexid
        self._name = name
        self._abbreviation = abbreviation
        self._country_id = country_id

        return

    indexid = property(lambda s: s._indexid)
    name = property(lambda s: s._name)
    abbreviation = property(lambda s: s._abbreviation)
    country_id = property(lambda s: s._country_id)



"""
Procuret Python
Unit of Account Module
author: hugh@blinkybeach.com
"""
from decimal import Decimal


class UnitOfAccount:
    """
    Abstract class defining a protocol for classes that represent units of
    account
    """

    exponent: int = NotImplemented

    def round_to_maximum_precision(self, amount: Decimal) -> Decimal:
        if not isinstance(self.exponent, int):
            raise NotImplementedError('Implement .exponent integer')
        return round(amount, self.exponent)

    def to_integer_smallest_unit(self, amount: Decimal) -> int:
        """
        Return an integer number of the smallest unit of this Unit that
        together comprise the supplied amount. For example, for USD 1,
        return 100.
        """
        return int(amount * (10**self.exponent))

    def from_integer_smallest_unit(self, amount: int) -> Decimal:
        """
        Return a Decimal magnitude value extracted from an integer
        representation of that value in the Unit's smallest possible
        denomination. For example, extract integer values from Stripe.
        """
        value = amount / Decimal(10**self.exponent)
        assert isinstance(value, Decimal)
        return value



"""
Procuret Python
Instalment Link Open Module
author: hugh@blinkybeach.com
"""
from procuret.time.time import ProcuretTime
from procuret.data.codable import Codable, CodingDefinition as CD
from typing import TypeVar, Type
from procuret.http.api_request import ApiRequest, HTTPMethod
from procuret.session import Session
from procuret.errors.type_error import ProcuretTypeError

Self = TypeVar('Self', bound='InstalmentLinkOpen')


class InstalmentLinkOpen(Codable):

    path = '/instalment-link/open'

    coding_map = {
        'sequence': CD(int),
        'created': CD(ProcuretTime)
    }

    def __init__(
        self,
        sequence: int,
        created: ProcuretTime
    ) -> None:

        self._sequence = sequence
        self._created = created

        return

    sequence = property(lambda s: s._sequence)
    created = property(lambda s: s._created)

    @classmethod
    def create(
        cls: Type[Self],
        link_id: str,
        session: Session
    ) -> None:

        if not isinstance(link_id, str):
            raise ProcuretTypeError('str', link_id, 'link_id')

        if not isinstance(session, Session):
            raise ProcuretTypeError('Session', session, 'session')

        _ = ApiRequest.make(
            path=cls.path,
            method=HTTPMethod.POST,
            data={'link_id': link_id},
            session=session
        )

        return None



"""
Procuret Python
Exercise Instalment Link Test
author: hugh@blinkybeach.com
"""
from procuret.tests.variants.with_supplier import TestWithSupplier
from procuret.tests.test_result import Success, TestResult
from procuret.instalment_link import InstalmentLink, InstalmentLinkOpen
from procuret.ancillary.communication_option import CommunicationOption
from decimal import Decimal
from procuret.session import Perspective


class ExerciseInstalmentLink(TestWithSupplier):

    NAME = 'Create, retrieve and mark an InstalmentLink as opened'

    test_perspective = Perspective.SUPPLIER

    def execute(self) -> TestResult:

        link = InstalmentLink.create(
            supplier=self.supplier_id,
            invoice_amount=Decimal('422.42'),
            invitee_email='noone@procuret-test-domain.org',
            invoice_identifier='Test 42',
            communication=CommunicationOption.EMAIL_CUSTOMER,
            session=self.session
        )

        assert isinstance(link, InstalmentLink)

        r_link = InstalmentLink.retrieve(
            public_id=link.public_id,
            session=self.session
        )

        assert isinstance(r_link, InstalmentLink)
        assert r_link.public_id == link.public_id

        InstalmentLinkOpen.create(
            link_id=r_link.public_id,
            session=self.session
        )

        o_link = InstalmentLink.retrieve(
            public_id=r_link.public_id,
            session=self.session
        )

        assert isinstance(o_link, InstalmentLink)
        assert len(o_link.opens) > 0
        assert isinstance(o_link.opens[0], InstalmentLinkOpen)

        return Success()



"""
Procuret Python
Test Suite Module
author: hugh@blinkybeach.com
"""
import sys
from procuret.tests import cases

STOP_AFTER_ONE = False
START = 0

if '--start' in sys.argv[1:] or '-s' in sys.argv[1:]:
    index = 0
    for key in sys.argv[1:]:
        if key == '--start' or key == '-s':
            START = int(sys.argv[1:][index + 1])
            break
        index += 1
        continue

if '--only-one' in sys.argv[1:] or '-o' in sys.argv[1:]:
    STOP_AFTER_ONE = True

LAST = False
if '--last' in sys.argv[1:]:
    LAST = True


TESTS = [
    cases.ExerciseInstalmentLink,
    cases.ExerciseTermRateGroup,
    cases.xero.RetrieveXeroEntityMap
]


def run_tests(start=START) -> None:
    """Execute Procuret Python unit tests"""
    print('Executing Procuret Python Test Suite')
    i = 0
    if LAST is True:
        start = len(TESTS)
    for test in TESTS:
        i += 1
        if i < start:
            continue
        number = str(i)
        while len(number) < 3:
            number = '0' + number
        number = '[' + number + '] '
        executed_test = test()
        executed_test.run()
        print(number + executed_test.report())
        if executed_test.did_fail:
            break
        if STOP_AFTER_ONE is True:
            break
    print('Procuret Python Test sequence complete.')
    return



"""
Procuret Python
Human Module
author: hugh@blinkybeach.com
"""
from enum import Enum
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.money.amount import Amount
from procuret.ancillary.entity_headline import EntityHeadline
from procuret.time.time import ProcuretTime
from procuret.payment_method.headline import PaymentMethodHeadline
from procuret.data.disposition import Disposition
from typing import TypeVar, Type, Optional, List
from procuret.data.order import Order
from procuret.session import Session
from procuret.http.api_request import ApiRequest, HTTPMethod, QueryParameters
from procuret.http.query_parameter import QueryParameter
from procuret.human_identity.identity import HumanIdentity
from procuret.human_identity_document.document import HumanIdentityDocument
from procuret.email_address.address import EmailAddress
from procuret.phone_number.number import PhoneNumber
from procuret.session import Session

Self = TypeVar('Self', bound='Human')


class Human(Codable):

    path = '/human'

    coding_map = {
        'public_id': CD(int),
        'agent_id': CD(int),
        'first_name': CD(str),
        'other_name': CD(str, optional=True),
        'last_name': CD(str),
        'identity': CD(HumanIdentity, optional=True),
        'documents': CD(
            HumanIdentityDocument,
            array=True,
            optional=True
        ),
        'created': CD(ProcuretTime, optional=True),
        'modified': CD(ProcuretTime, optional=True),
        'email_address': CD(EmailAddress),
        'phone_number': CD(PhoneNumber),
        'disposition': CD(Disposition)
    }
    
    def __init__(
        self,
        public_id: int,
        agent_id: int,
        first_name: str,
        other_name: Optional[str],
        last_name: str,
        identity: Optional[HumanIdentity],
        documents: Optional[List[HumanIdentityDocument]],
        created: Optional[ProcuretTime],
        modified: Optional[ProcuretTime],
        email_address: EmailAddress,
        phone_number: PhoneNumber,
        disposition: Disposition
    ) -> None:

        self._public_id = public_id
        self._agent_id = agent_id
        self._first_name = first_name
        self._other_name = other_name
        self._last_name = last_name
        self._identity = identity
        self._documents = documents
        self._created = created
        self._modified = modified
        self._email_address = email_address
        self._phone_number = phone_number
        self._disposition = disposition

        return

    public_id: int = property(lambda s: s._public_id)
    agent_id: int = property(lambda s: s._agent_id)
    first_name: str = property(lambda s: s._first_name)
    other_name: Optional[str] = property(lambda s: s._other_name)
    last_name: str = property(lambda s: s._last_name)
    identity: Optional[HumanIdentity] = property(lambda s: s._identity)
    documents: List[HumanIdentityDocument] = property(
        lambda s: s._documents or []
    )
    created: Optional[ProcuretTime] = property(lambda s: s._created)
    modified: Optional[ProcuretTime] = property(lambda s: s._modified)
    email_address: EmailAddress = property(lambda s: s._email_address)
    phone_number: PhoneNumber = property(lambda s: s._phone_number)
    disposition: Disposition = property(lambda s: s._disposition)

    @classmethod
    def retrieve(
        Self: Type[Self],
        public_id: str,
        session: Session
    ) -> Optional[Self]:

        result = ApiRequest.make(
            path=Self.path,
            method=HTTPMethod.GET,
            data=None,
            session=session,
            query_parameters=QueryParameters([
                QueryParameter('human_id', public_id)
            ])
        )

        return Self.optionally_decode(result)



"""
Procuret Python
Instalment Total Row Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from decimal import Decimal
from procuret.money.currency import Currency


class InstalmentTotalRow(Codable):

    coding_map = {
        'total_payments': CD(Decimal),
        'total_principal': CD(Decimal),
        'total_principal_discount': CD(Decimal),
        'total_interest': CD(Decimal),
        'denomination': CD(Currency)
    }

    def __init__(
        self,
        total_payments: Decimal,
        total_principal: Decimal,
        total_principal_discount: Decimal,
        total_interest: Decimal,
        denomination: Currency
    ) -> None:

        self._total_payments = total_payments
        self._total_principal = total_principal
        self._total_principal_discount = total_principal_discount
        self._total_interest = total_interest
        self._denomination = denomination

        return
    
    total_payments: Decimal = property(lambda s: s._total_payments)
    total_principal: Decimal = property(lambda s: s._decimal)
    total_principal_discount: Decimal = property(
        lambda s: s._total_principal_discount
    )
    total_interest: Decimal = property(lambda s: s._total_interest)
    denomination: Currency = property(lambda s: s._denomination)



"""
Procuret Python
Commitment Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.human.human import Human
from typing import TypeVar, Type, Optional
from procuret.session import Session
from procuret.http.api_request import ApiRequest, HTTPMethod, QueryParameters
from procuret.http.query_parameter import QueryParameter

Self = TypeVar('Self', bound='Commitment')


class Commitment(Codable):

    path = '/commitment'

    coding_map = {
        'public_id': CD(str),
        'committed_human': CD(Human)
    }

    def __init__(
        self,
        public_id: str,
        committed_human: Human
    ) -> None:

        self._public_id = public_id
        self._committed_human = committed_human

        return

    public_id = property(lambda s: s._public_id)
    committed_human = property(lambda s: s._committed_human)

    @classmethod
    def retrieve(
        Self: Type[Self],
        exchange_id: str,
        session: Session
    ) -> Optional[Self]:

        result = ApiRequest.make(
            path=Self.path,
            method=HTTPMethod.GET,
            data=None,
            session=session,
            query_parameters=QueryParameters([QueryParameter(
                key='setup_id',
                value=exchange_id
            )])
        )

        return Self.optionally_decode(result)



"""
Procuret API
Entity Identifier Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.entity_id_type.type import EntityIdType

class EntityIdentifier(Codable):

    coding_map = {
        'id': CD(str),
        'id_type': CD(EntityIdType)
    }

    def __init__(
        self,
        id: str,
        id_type: EntityIdType
    ) -> None:

        self._id = id
        self._id_type = id_type

        return
    
    identifier: str = property(lambda s: s._id)
    id_type: EntityIdType = property(lambda s: s._entity_id_type)



"""
Procuret Python
Instalment Schedule Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from typing import List, Any, TypeVar, Type, Optional
from procuret.instalment_line.line import InstalmentLine
from procuret.instalment_total_row.row import InstalmentTotalRow
from procuret.session import Session
from procuret.http.api_request import ApiRequest, HTTPMethod, QueryParameters
from procuret.http.query_parameter import QueryParameter

Self = TypeVar('Self', bound='InstalmentSchedule')


class InstalmentSchedule(Codable):

    coding_map = {
        'lines': CD(InstalmentLine, array=True),
        'series_id': CD(str),
        'has_interest': CD(bool),
        'total_row': CD(InstalmentTotalRow)
    }

    def __init__(
        self,
        lines: List[InstalmentLine],
        series_id: str,
        has_interest: bool,
        total_row: InstalmentTotalRow
    ) -> None:

        self._lines = lines
        self._series_id = series_id
        self._has_interest = has_interest
        self._total_row = total_row

        return

    lines: List[InstalmentLine] = property(lambda s: s._lines)
    series_id: str = property(lambda s: s._series_id)
    has_interest: bool = property(lambda s: s._has_interest)
    total_row: InstalmentTotalRow = property(lambda s: s._total_row)

    @classmethod
    def retrieve(
        Self: Type[Self],
        series_id: str,
        session: Session
    ) -> Optional[Self]:

        result = ApiRequest.make(
            path=Self.path,
            method=HTTPMethod.GET,
            data=None,
            session=session,
            query_parameters=QueryParameters([QueryParameter(
                key='entity_id',
                value=series_id
            )])
        )

        return Self.optionally_decode(result)



"""
Procuret Python
Test Result Module
author: hugh@blinkybeach.com
"""
from typing import Optional


class TestResult:
    """Abstract class defining an interface for test result classes"""
    DID_PASS: bool = NotImplemented


class Success(TestResult):
    DID_PASS = True


class Failure(TestResult):
    DID_PASS = False

    def __init__(
        self,
        description: str,
        exception: Optional[Exception] = None
    ) -> None:

        self._description = description
        self._exception = exception

        return

    description = property(lambda s: s._description)
    exception = property(lambda s: s._exception)



"""
Procuret Python
Entity Headline Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD


class EntityHeadline(Codable):

    coding_map = {
        'legal_entity_name': CD(str),
        'entity_id': CD(int)
    }

    def __init__(
        self,
        legal_entity_name: str,
        entity_id: int
    ) -> None:

        self._legal_entity_name = legal_entity_name
        self._entity_id = entity_id

        return

    legal_entity_name = property(lambda s: s._legal_entity_name)
    entity_id = property(lambda s: s._entity_id)



"""
Procuret Python
Term Rate Group Module
author: hugh@blinkybeach.com
"""
from procuret.data.order import Order
from procuret.data.codable import Codable, CodingDefinition as CD
from enum import Enum
from typing import Optional, List, TypeVar, Type
from procuret.time.time import ProcuretTime
from procuret.data.disposition import Disposition
from procuret.term_rate.term_rate import TermRate
from procuret.http.api_request import ApiRequest, HTTPMethod
from procuret.http.query_parameters import QueryParameter, QueryParameters
from procuret.errors.type_error import ProcuretTypeError
from procuret.session import Session

Self = TypeVar('Self', bound='TermRateGroup')


class OrderBy(Enum):
    CREATED = 'created'


class TermRateGroup(Codable):

    path = TermRate.path + '/group'
    list_path = path + '/list'

    coding_map = {
        'public_id': CD(str),
        'rates': CD(TermRate, array=True),
        'created': CD(ProcuretTime),
        'active': CD(bool),
        'disposition': CD(Disposition, optional=True),
        'order': CD(Order),
        'order_by': CD(OrderBy)
    }

    def __init__(
        self,
        public_id: str,
        rates: List[TermRate],
        created: ProcuretTime,
        active: bool,
        disposition: Optional[Disposition],
        order: Order,
        order_by: OrderBy
    ) -> None:

        self._public_id = public_id
        self._rates = rates
        self._created = created
        self._active = active
        self._disposition = disposition
        self._order = order
        self._order_by = order_by

        return

    public_id = property(lambda s: s._public_id)
    rates = property(lambda s: s._rates)
    created = property(lambda s: s._created)
    active = property(lambda s: s._active)
    disposition = property(lambda s: s._disposition)
    order = property(lambda s: s._order)
    order_by = property(lambda s: s._order_by)

    @classmethod
    def retrieve(
        cls: Type[Self],
        public_id: str,
        session: Session
    ) -> Optional[Self]:

        if not isinstance(public_id, str):
            raise ProcuretTypeError('str', public_id, 'public_id')

        if not isinstance(session, Session):
            raise ProcuretTypeError('Session', session, 'session')

        parameters = [QueryParameter('public_id', public_id)]

        result = ApiRequest.make(
            path=cls.path,
            method=HTTPMethod.GET,
            data=None,
            session=session,
            query_parameters=QueryParameters(parameters)
        )

        return cls.optionally_decode(result)



"""
Procuret Python
Term Rate Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from decimal import Decimal


class TermRate(Codable):

    path = '/term-rate'

    coding_map = {
        'supplier_entity_id': CD(int),
        'periods': CD(int),
        'periods_in_year': CD(Decimal)
    }

    def __init__(
        self,
        supplier_entity_id: int,
        periods: int,
        periods_in_year: Decimal
    ) -> None:

        self._supplier_entity_id = supplier_entity_id
        self._periods = periods
        self._periods_in_year = periods_in_year

        return

    periods = property(lambda s: s._periods)
    periods_in_year = property(lambda s: s._periods_in_year)
    supplier_entity_id = property(lambda s: s._supplier_entity_id)



"""
Procuret Python
Session Module
author: hugh@blinkybeach.com
"""
import json
from procuret.data.codable import CodingDefinition as CD
from procuret.ancillary.abstract_session import AbstractSession
from procuret.ancillary.session_lifecycle import Lifecycle
from procuret.ancillary.session_perspective import Perspective
from procuret.http.api_request import ApiRequest
from procuret.http.method import HTTPMethod
from typing import TypeVar, Type, Optional
from procuret.errors.inconsistent import InconsistentState
from procuret.security.second_factor_code import SecondFactorCode
from getpass import getpass
from base64 import b64decode, b64encode

Self = TypeVar('Self', bound='Session')


class Session(AbstractSession):

    PATH = '/session'

    coding_map = {
        'session_id': CD(int),
        'session_key': CD(str),
        'api_key': CD(str),
        'lifecycle': CD(Lifecycle),
        'perspective': CD(Perspective)
    }

    def __init__(
        self,
        session_id: int,
        session_key: str,
        api_key: str,
        lifecycle: Lifecycle,
        perspective: Perspective,
        on_behalf_of: Optional[int] = None
    ) -> None:

        self._session_id = session_id
        self._session_key = session_key
        self._api_key = api_key
        self._lifecycle = lifecycle
        self._perspective = perspective
        self._on_behalf_of = on_behalf_of

        return

    session_id = property(lambda s: s._session_id)
    session_key = property(lambda s: s._session_key)
    api_key = property(lambda s: s._api_key)
    lifecycle = property(lambda s: s._lifecycle)
    perspective = property(lambda s: s._perspective)
    on_behalf_of = property(lambda s: s._on_behalf_of)

    acts_for_another_agent = property(lambda s: s.on_behalf_of is not None)

    def _to_storage_format(self) -> str:
        return b64encode(
            json.dumps(self.encode()).encode('utf-8')
        ).decode('utf-8')


    def save_to_file(
        self,
        filepath: str
    ) -> None:

        with open(filepath, 'w') as wfile:
            wfile.write(self._to_storage_format())

        return None

    def acting_for(self, agent: int) -> Self:
        return Session(
            session_id=self._session_id,
            session_key=self._session_key,
            api_key=self._api_key,
            lifecycle=self._lifecycle,
            perspective=self._perspective,
            on_behalf_of=agent
        )

    @classmethod
    def load_from_file(
        self,
        filepath: str
    ) -> Self:

        with open(filepath, 'r') as rfile:
            return Session._from_storage_format(rfile.read())

    @classmethod
    def create_with_email(
        Self: Type[Self],
        email: str,
        plaintext_secret: str,
        perspective: Perspective,
        code: str,
        lifecycle: Lifecycle = Lifecycle.LONG_LIVED
    ) -> Self:

        data = {
            'email': email,
            'secret': plaintext_secret,
            'perspective': perspective.value,
            'lifecycle': lifecycle.value,
            'code': code
        }

        result = ApiRequest.make(
            path=Self.PATH,
            method=HTTPMethod.POST,
            data=data,
            session=None,
            query_parameters=None
        )
        if result is None:
            raise InconsistentState

        return Self.decode(result)

    def interactively_to_file(
        self
    ) -> None:

        print('Enter the filepath into which you with to save this Session.')
        filepath = input('filepath: ')

        self.save_to_file(filepath)

        print(f'This Session has been saved to {filepath}')

        return None

    @classmethod
    def from_interactive_prompt(
        Self: Type[Self]
    ) -> Self:

        print('\nInteractive Procuret Session Creation Prompt\n')
        print('*** Be careful with Sessions! Never store them where someone el\
se could access them. Never give them to anyone who asks, even if they say the\
y are from Procuret. Always keep your passphrase a secret. ***')

        print("""
What perspective would you like the Session to have?
"1" for supplier
"2" for business (Customer)
"3" for administrator
"4" for sales (PM)
"5" for treasury
"6" for investor
"7" for customer_support (Customer Success)
""")
        perspective = Perspective(int(input('perspective: ')))

        print("""
Would you like the Session to be long lived, or short lived? Choose long-lived
if the Session will be used for a very long time. If unsure, choose
short-lived.
"1" for short lived
"2" for long lived
""")
        lifecycle = Lifecycle(int(input('lifecycle: ')))

        print("Now, input your Procuret account credentials.")
        email_address = input('email: ')
        secret = getpass('password: ')

        print('Requesting second factor code...')

        SecondFactorCode.create_with_email(
            email=email_address,
            plaintext_secret=secret,
            perspective=perspective
        )

        print(f'Enter the six digit code Procuret just sent to \
{email_address}')

        code = input('code: ')

        print('Creating Session...')
    
        session = Self.create_with_email(
            email=email_address,
            plaintext_secret=secret,
            perspective=perspective,
            lifecycle=lifecycle,
            code=code
        )

        print(f'Session created with ID {session._session_id}')

        return session

    @classmethod
    def _from_storage_format(
        Self: Type[Self],
        data: str
    ) -> Self:

        return Self.decode(json.loads(b64decode(data).decode('utf-8')))



"""
Procuret Python
Human Identity Document State Module
author: hugh@blinkybeach.com
"""
from enum import Enum


class HumanIdentityDocumentState(Enum):
    NOT_VERIFIED = 1
    VERIFIED_VALID = 2
    INVALID = 3



"""
Procuret Python
Communication Option Module
author: hugh@blinkybeach.com
"""
from enum import Enum


class CommunicationOption(Enum):
    EMAIL_CUSTOMER = 'email_customer'
    DO_NOT_CONTACT_CUSTOMER = 'do_not_contact'



"""
Procuret Python
Xero Entity Map Module
author: hugh@blinkybeach.com
"""
from procuret.errors.type_error import ProcuretTypeError
from typing import Optional, TypeVar, Type
from procuret.http.query_parameters import QueryParameter, QueryParameters
from procuret.data.codable import CodingDefinition as CD, Codable
from procuret.http.api_request import ApiRequest, HTTPMethod
from procuret.session import Session

Self = TypeVar('Self', bound='XeroEntityMap')


class XeroEntityMap(Codable):

    path = '/xero/entity-map'

    coding_map = {
        'entity_id': CD(int)
    }

    def __init__(
        self,
        entity_id: int
    ) -> None:

        self._entity_id = entity_id

        return

    entity_id = property(lambda s: s._entity_id)

    @classmethod
    def retrieve(
        cls: Type[Self],
        entity_id: int,
        session: Session
    ) -> Optional[Self]:

        if not isinstance(entity_id, int):
            raise ProcuretTypeError('int', entity_id, 'entity_id')

        result = ApiRequest.make(
            path=cls.path,
            method=HTTPMethod.GET,
            data=None,
            session=session,
            query_parameters=QueryParameters(targets=[
                QueryParameter('entity_id', entity_id)
            ])
        )

        return cls.optionally_decode(result)



"""
Procuret Python
Global Brand Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.global_brand.brand import GlobalBrand
from procuret.time.time import ProcuretTime


class GlobalBrandSelection(Codable):

    coding_map = {
        'created': CD(ProcuretTime),
        'brand': CD(GlobalBrand)
    }

    def __init__(
        self,
        created: ProcuretTime,
        brand: GlobalBrand
    ) -> None:

        self._created = created
        self._brand = brand

        return

    created = property(lambda s: s._created)
    brand = property(lambda s: s._brand)



"""
Procuret Python
Address Module
author: hugh@blinkybeach.com
"""
from typing import Optional
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.time.time import ProcuretTime
from procuret.region.region import Region
from procuret.country.country import Country


class Address(Codable):

    coding_map = {
        'created': CD(ProcuretTime),
        'line_1': CD(str),
        'line_2': CD(str, optional=True),
        'line_3': CD(str, optional=True),
        'line_4': CD(str, optional=True),
        'locality': CD(str),
        'region': CD(Region, optional=True),
        'postal_code': CD(str),
        'country': CD(Country, optional=True)
    }

    def __init__(
        self,
        created: ProcuretTime,
        line_1: str,
        line_2: Optional[str],
        line_3: Optional[str],
        line_4: Optional[str],
        locality: str,
        region: Optional[Region],
        postal_code: str,
        country: Optional[Country]
    ) -> None:

        self._created = created
        self._line_1 = line_1
        self._line_2 = line_2
        self._line_3 = line_3
        self._line_4 = line_4
        self._locality = locality
        self._region = region
        self._postal_code = postal_code
        self._country = country

        return

    created = property(lambda s: s._created)
    line_1 = property(lambda s: s._line_1)
    line_2 = property(lambda s: s._line_2)
    line_3 = property(lambda s: s._line_3)
    line_4 = property(lambda s: s._line_4)
    locality = property(lambda s: s._locality)
    region = property(lambda s: s._region)
    postal_code = property(lambda s: s._postal_code)
    country = property(lambda s: s._country)

    one_line = property(lambda s: s._in_one_line())

    def _in_one_line(self) -> str:
        """Return this Address a single line string"""
        line = self._line_1.value
        if self._line_2:
            line += ', ' + self._line_2.value
        if self._line_3:
            line += ', ' + self._line_3.value
        if self._line_4:
            line += ', ' + self._line_4.value
        if self._locality:
            line += ', ' + self._locality
        if self._region:
            line += ' ' + self._region.abbreviation
        line += ' ' + self._postal_code
        return line + ' ' + self._country.iso_3166_a3



"""
Procuret Python
Api Request Module
author: hugh@blinkybeach.com
"""
import json
import os.path
from procuret.ancillary.command_line import CommandLine
from procuret.http.method import HTTPMethod
from typing import TypeVar, Type, Optional, Dict, Union
from procuret.data.codable import Object, Array
from procuret.http.query_parameters import QueryParameters
from procuret.ancillary.abstract_session import AbstractSession
from procuret.ancillary.session_lifecycle import Lifecycle
from procuret.errors.api_error import ProcuretApiError
from datetime import datetime
import hmac
from hashlib import sha256
from base64 import b64encode
from urllib.request import HTTPError
from urllib.request import Request
from urllib.request import urlopen
from procuret.version import VERSION as AGENT_VERSION
import sys

T = TypeVar('T', bound='ApiRequest')

API_ENDPOINT = 'https://procuret.com/api'

CL = CommandLine.load()
OVERRIDE_ENDPOINT = CL.get(key='--procuret-api-endpoint')
if OVERRIDE_ENDPOINT is not None:
    API_ENDPOINT = OVERRIDE_ENDPOINT

_AGENT_PREFIX: Optional[str] = None
if os.path.exists('procuret_configuration'):
    with open('procuret_configuration', 'r') as rfile:
        json_data = json.loads(rfile.read())
        API_ENDPOINT = json_data['api_endpoint']
        _AGENT_PREFIX = json_data['agent_prefix'] if (
            'agent_prefix' in json_data.keys()
        ) else None
    pass

def _make_user_agent() -> str:
    base = f'Procuret Python {AGENT_VERSION};Python {sys.version}'
    if _AGENT_PREFIX is not None:
        return f'{_AGENT_PREFIX};{base}'
    return base

class ApiRequest:

    API_KEY_NAME = 'x-procuret-api-key'
    SIGNATURE_KEY_NAME = 'x-procuret-signature'
    SESSION_ID_NAME = 'x-procuret-session-id'
    USER_AGENT = 'Procuret Python ' + AGENT_VERSION
    FORWARDED_AGENT = 'x-procuret-forwarded-agent'

    @classmethod
    def make(
        cls,
        path: str,
        method: HTTPMethod,
        data: Optional[Union[Object, Array]],
        session: Optional[AbstractSession],
        query_parameters: Optional[QueryParameters] = None,
        api_endpoint: str = API_ENDPOINT
    ) -> Optional[Union[Object, Array]]:
        """Return the decoded json body of a response from the Procuret API"""

        url = api_endpoint + path
        if query_parameters is not None:
            url = query_parameters.add_to(url)

        headers = {'User-Agent': _make_user_agent()}

        if session is not None:
            headers = cls._add_authorisation_to_headers(
                headers=headers,
                path=path,
                session=session
            )

        if session is not None and session.acts_for_another_agent:
            headers = cls._add_forwarded_agent_to_headers(
                headers=headers,
                session=session
            )

        encoded_data: Optional[bytes] = None
        if data is not None:
            encoded_data = json.dumps(data).encode('utf-8')
            headers['Content-Type'] = 'application/json'

        request = Request(
            url=url,
            method=method.value,
            data=encoded_data,
            headers=headers
        )

        try:
            response = urlopen(request).read()
        except HTTPError as error:
            if error.code == 404:
                return None
            raise ProcuretApiError(error.code)

        return json.loads(response)

    @classmethod
    def _add_authorisation_to_headers(
        cls: Type[T],
        headers: Dict[str, str],
        path: str,
        session: AbstractSession
    ) -> Dict[str, str]:

        headers[cls.SESSION_ID_NAME] = str(session.session_id)

        if session.lifecycle == Lifecycle.SHORT_LIVED:
            return cls._headers_with_key(
                headers=headers,
                api_key=session.api_key
            )

        if session.lifecycle == Lifecycle.LONG_LIVED:
            return cls._headers_with_signature(
                headers=headers,
                api_key=session.api_key,
                path=path
            )

        raise NotImplementedError

    @classmethod
    def _add_forwarded_agent_to_headers(
        cls: Type[T],
        headers: Dict[str, str],
        session: AbstractSession
    ) -> Dict[str, str]:

        if not session.acts_for_another_agent:
            return headers

        headers[cls.FORWARDED_AGENT] = str(session.on_behalf_of)

        return headers

    @classmethod
    def _headers_with_key(
        cls: Type[T],
        headers: Dict[str, str],
        api_key: str
    ) -> Dict[str, str]:

        headers[cls.API_KEY_NAME] = api_key
        return headers

    @classmethod
    def _headers_with_signature(
        cls: Type[T],
        headers: Dict[str, str],
        api_key: str,
        path: str
    ) -> Dict[str, str]:

        time = int(datetime.now().timestamp())
        payload = (str(time - (time % 900)) + path).encode('utf-8')
        digest = hmac.new(api_key.encode('utf-8'), payload, sha256).digest()

        headers[cls.SIGNATURE_KEY_NAME] = b64encode(digest).decode()

        return headers



"""
Procuret API
Entity Module
author: hugh@blinkybeach.com
"""
from procuret.data.codable import Codable, CodingDefinition as CD
from procuret.data.disposition import Disposition
from typing import List, TypeVar, Type, Optional
from procuret.session import Session
from procuret.entity_identifier.identifier import EntityIdentifier
from procuret.http.api_request import ApiRequest, HTTPMethod, QueryParameters
from procuret.http.query_parameter import QueryParameter
from procuret.address.address import Address

Self = TypeVar('Self', bound='Entity')


class Entity(Codable):

    path = '/entity'

    coding_map = {
        'public_id': CD(int),
        'public_id_short': CD(str),
        'entity_identifiers': CD(EntityIdentifier, array=True),
        'legal_entity_name': CD(str),
        'business_address': CD(Address),
        'disposition': CD(Disposition)
    }

    def __init__(
        self,
        public_id: str,
        public_id_short: str,
        entity_identifiers: List[EntityIdentifier],
        legal_entity_name: str,
        business_address: Address,
        disposition: Disposition
    ) -> None:

        self._public_id = public_id
        self._public_id_short = public_id_short
        self._entity_identifiers = entity_identifiers
        self._legal_entity_name = legal_entity_name
        self._business_address = business_address
        self._disposition = disposition

        return

    public_id: int = property(lambda s: s._public_id)
    public_id_short: str = property(lambda s: s._public_id_short)
    entity_identifiers: List[EntityIdentifier] = property(
        lambda s: s._entity_identifiers
    )
    legal_entity_name: str = property(lambda s: s._legal_entity_name)
    disposition: Disposition = property(lambda s: s._disposition)
    address: Address = property(lambda s: s._business_address)

    @classmethod
    def retrieve(
        Self: Type[Self],
        public_id: int,
        session: Session
    ) -> Optional[Self]:

        result = ApiRequest.make(
            path=Self.path,
            method=HTTPMethod.GET,
            data=None,
            session=session,
            query_parameters=QueryParameters([QueryParameter(
                key='entity_id',
                value=public_id
            )])
        )

        return Self.optionally_decode(result)



from procuret.instalment_link.instalment_link import InstalmentLink
from procuret.instalment_link.open import InstalmentLinkOpen
from procuret.instalment_link.instalment_link import (
    OrderBy as InstalmentLinkOrderBy
)
